// worst case stack pointer usage (call tree) generator
// based in part on previous work by "mammon_/the_owl/slava" (public domain, from the internet)
//
// History:
// __before__  twh
//             - initial coding and testing
// 10.27.2000  twh
//             - Added GetEndSP() to account for $sp_regsv$3 and other similar library calls
// 01.22.2001  twh
//             - Travelunc() now correctly handles JMP & BRA to code outside of the current function
//             - OutputLine() split into OutputLineJMP() and OutputLineJSR()
// 03.08.2001  twh
//             - corrected small bug in GetEndSP() where computed ending address was sometimes incorrectly offset by two bytes
//             - OutputLineJSR() and OutputLineJMP() merged back into OutputLine() with extra parameter
//             - TravelFunc() now outputs depth first, rather then in order
//             - GetInputPath() more optimized
// 03.09.2001  twh
//             - main() now prompts for an initial SP offset value
//             - TravelFunc() simplified & renamed to VisitFunc()
//             - removed GetEndSP() and GetMaxSP() as they are not being used anymore
//             - simplified and documented more code throughout
// 03.09.2001  saf/twh
//             - verified by HDI E6000 trace that results match images
//             - verified by hand trace (w/ IDA Pro & pencil & paper) images
// 03.12.2001  twh
//             - final documentation

#include <idc.idc>

#define TRUE    1
#define FALSE   0


// output a single line of the call tree / stack pointer usage analysis
static OutputLine(ea, x, nest, OutFileH, ss, ct) {
    auto j;

    // output call depth and max SP used for the current subroutine
    fprintf(OutFileH, "%s\t%s\t", ltoa(nest,10), ltoa(ss,10) );

    // use tabs to indent the current subroutine name
    for ( j = 0; j < nest; j = j + 1)
        fprintf(OutFileH, "\t");

    // output the calling address and the destination subroutine
    fprintf(OutFileH, "%s  %s>  %s : %s\n", atoa(ea), ct, Name(x), atoa(x) );
}


// recursive function to iterate through all subroutine calls from a given address
static VisitFunc(oea, nea, nest, MaxNest, ArrayID, OutFileH, spIn, cjt) {
    auto eaFend, eaBranch;
    auto i;
    auto spCur, spMax, spDif;
    auto flType, ct, co, fl;
    auto ea, pea;

    // recursive depth guard check
    if (nest >= (MaxNest-1))
        return 0;

    ea = nea;
    spMax = spIn;
    spCur = spIn;

    // note that FindFuncEnd returns the address of (FunctionEnd+1)
    eaFend = FindFuncEnd(ea);

    // search entire function for calls
    for ( pea = BADADDR; ea < eaFend; ea=NextAddr(ea) )
    {
        if (pea == BADADDR)
            spCur = spIn - GetSpd(ea);
        else
            spCur = spCur - GetSpd(ea) + GetSpd(pea);
        if (spCur > spMax)
            spMax = spCur;
        pea = ea;

        eaBranch = Rfirst0(ea);
        while (eaBranch != BADADDR)
        {
            // call far or call near ?
            flType = XrefType();

            ct = "-";
            co = 2;
            fl = FALSE;
            // handle JMP, BRA, etc...  (0 SP offset)
            if ( ( (flType == fl_JF) || (flType == fl_JN) ) && (GetFunctionName(eaBranch) !=  GetFunctionName(ea)) )
            {
                ct = "=";
                co = 0;
                fl = TRUE;
            }
            // handle JSR, BSR, etc...  (2 SP offset)
            if ( (flType == fl_CF) || (flType == fl_CN) )       // || (flType == fl_US)    for user specified types
                fl = TRUE;

            if (fl)
            {
                // check if the current function is part of the current tree already
                for (i = 0; i <= GetLastIndex(AR_STR, ArrayID); i = i + 1) {
                    if (Name(eaBranch) == GetArrayElement(AR_STR, ArrayID, i)) {
                        fprintf(OutFileH, "***** RECURSIVE CALL OF FUNCTION   %s *****\n", Name(eaBranch) );
                        DelArrayElement(AR_STR, ArrayID, nest+1);
                        return 0;
                    }
                }

                // add current function to array
                SetArrayString(ArrayID, nest+1, Name(eaBranch));

				//fprintf(OutFileH, "%s: %s\n", ltoa(ea,16), ltoa(spCur,10) );
                // the recursive call
                spDif = VisitFunc(ea, eaBranch, nest+1, MaxNest, ArrayID, OutFileH, spCur + co, ct);

				//fprintf(OutFileH, "%s: %s, %s\n", ltoa(ea,16), ltoa(spCur,10), ltoa(spDif,10) );

                // adjust SP by offset for "skewed" funcs...
                spCur = spCur + spDif;

                DelArrayElement(AR_STR, ArrayID, nest+1);
            }
            eaBranch = Rnext0(ea, eaBranch);
        }
    }
    // display current data
    OutputLine(oea, nea, nest+1, OutFileH, spMax, cjt);

    return spCur - spIn;
}


static GetInputPath()
{
    auto p, r, e;

    p = GetInputFilePath();
    r = -1;
    e = strlen( p );
    while ( r < 0 )
        r = strstr(substr(p, --e, -1), "\\");

    return( substr(p, 0, ++e) );
}


// main deals with the interface presented to the user and then performs the analysis
static main()
{
    auto ea;
    auto ArrayID;
    auto OutFileH, OutFName;
    auto nest, MaxNest;
    auto spBegin, spEnd;

    // create an array to hold function call addresses (to detect/prevent recursion)
    ArrayID =  CreateArray("MyFuncAddrs");
    if (ArrayID == -1)
    {
        ArrayID = GetArrayId("MyFuncAddrs");
        DeleteArray(ArrayID);
        ArrayID =  CreateArray("MyFuncAddrs");
        if (ArrayID == -1)
        {
            Warning("Cannot create an array to store function addresses, exiting!");
            return;
        }
    }

    ea = ChooseFunction("Select a function to examine:");
    nest = -1;
    spBegin = atol( AskStr( "0", "Enter the initial SP offset:") );
    MaxNest = atol( AskStr( "64", "Enter the maximum call depth to traverse:") );
    OutFName = AskFile(1, GetInputPath() + Name(ea) + ".txt", "Enter output text filename: ");
    OutFileH = fopen(OutFName, "wt");

    fprintf(OutFileH, "Depth\tSP\tLocation\n");

    // No globals allowed.  Must pass everything.  Sigh...
    spEnd = VisitFunc(ea, ea, nest, MaxNest, ArrayID, OutFileH, spBegin, "=");
    Message("ending offset is %s\n", ltoa(spEnd,10) );

    fclose (OutFileH);
    DeleteArray(ArrayID);
    Message("End of output.\n");
}

